#include <sys/farptr.h>
#include <go32.h>
#include <dpmi.h>
#include <string.h>
#include <pc.h>
#include <stdlib.h>
#include <stdio.h>


typedef unsigned char       BYTE;
typedef unsigned short      WORD;
typedef unsigned long       DWORD;
typedef int BOOL;

#ifndef FALSE
#define FALSE   0
#endif

#ifndef TRUE
#define TRUE    !FALSE
#endif



//----- YAMAHA DS-XG Devices -----------------------------------------------
#define YAMAHA      0x1073
#define YMF724      0x0004
#define YMF724F     0x000d
#define YMF734      0x0005
#define YMF737      0x0008
#define YMF738      0x0020
#define YMF740      0x000a
#define YMF740C     0x000c
#define YMF744      0x0010
#define YMF754      0x0012
#define YMF738_TEG  0x0006


static BOOL YMFIRQ(int forceirq)
{
    __dpmi_regs r;
    int index;
    DWORD device_id[] = {
        YMF724,
        YMF724F,
        YMF734,
        YMF737,
        YMF738,
        YMF740,
        YMF740C,
        YMF744,
        YMF754,
        YMF738_TEG,
        0xFFFFFFFF
    };
    char device_name[][16] = {
        "YMF724",
        "YMF724F",
        "YMF734",
        "YMF737",
        "YMF738",
        "YMF740",
        "YMF740C",
        "YMF744",
        "YMF754",
        "YMF738_TEG"
    };

    r.d.eax = 0x0000b101;                       // PCI BIOS - INSTALLATION CHECK
    r.d.edi = 0x00000000;
    __dpmi_int(0x1a, &r);
    if( r.d.edx != 0x20494350 ){                // ' ICP'
        printf("no pci bios found\n");
        return 0;
    }

    // find device
    index = 0;
    while(device_id[index] != 0xFFFFFFFF){
        r.d.eax = 0x0000B102;                   // PCI BIOS - FIND PCI DEVICE
        r.d.esi = 0x00000000;                   // device index
        r.d.edx = YAMAHA;                       // vendor ID
        r.d.ecx = device_id[index];              // device ID
        __dpmi_int(0x1a, &r);
        if( r.h.ah == 0 ){
            printf("%s found.", device_name[index]);
            break;
        }
        index += 1;
    }

    if(device_id[index] == 0xFFFFFFFF){
        printf("no ymfpci found\n");
        return FALSE;
    }

    // set irq
    r.d.eax = 0x0000B10F;                   // SET PCI HARDWARE INTERRUPT
                                            // BH = bus number
                                            // BL = device/function number
    r.h.cl  = 0x0A;
    r.h.ch  = forceirq & 0x0F;
    r.x.ds  = 0xF000;
    __dpmi_int(0x1a, &r);
    if( r.h.ah != 0 ){
        goto error;
    }

    r.d.eax = 0x0000B10A;                   // READ CONFIGURATION DWORD
                                            // BH = bus number
                                            // BL = device/function number
    r.d.edi = 0x0000003C;                   // register number
    __dpmi_int(0x1a, &r);
    if( r.h.ah != 0 ){
        goto error;
    }
    r.d.ecx = r.d.ecx & 0xFFFFFF00;
    r.d.ecx = r.d.ecx | (forceirq & 0x0F);

    r.d.eax = 0x0000B10C;                   // Write CONFIGURATION DWORD
                                            // BH = bus number
                                            // BL = device/function number
    r.d.edi = 0x0000003C;                   // register number
    __dpmi_int(0x1a, &r);
    if( r.h.ah != 0 ){
        goto error;
    }


    r.d.eax = 0x0000B10A;                       // READ CONFIGURATION DWORD
                                                // BH = bus number
                                                // BL = device/function number
    r.d.edi = 0x0000003C;                       // register number
    __dpmi_int(0x1a, &r);
    if( r.h.ah != 0 ){
        goto error;
    }
    printf(" IRQ was changed into %d\n", r.d.ecx & 0x0F);

    return TRUE;

error:
    printf("\nerror\n");
    return FALSE;
}


static void set_irq_edge_or_level(int irq, int edge_or_level)
{
    /*
     *  edge_or_level == 0     set edge  triggered
     *                != 0     set level triggered
     */

    int bitmap;

    if(irq > 15) return;
    if(edge_or_level != 0) edge_or_level = 1;
    if(irq < 8){
        bitmap = inportb(0x04D0);
        bitmap = bitmap & ~(1 << irq);
        bitmap = bitmap | (edge_or_level << irq);
        outportb(0x04D0, bitmap);
    }else{
        irq -= 8;
        bitmap = inportb(0x04D1);
        bitmap = bitmap & ~(1 << irq);
        bitmap = bitmap | (edge_or_level << irq);
        outportb(0x04D1, bitmap);
    }
}



static void help(void)
{
    printf("usage:  ymfirq  irq_number\n");
}




int main(int argc, char* argv[])
{
    int index;
    int force_irq  = 5;
    int trigger_mode = -1;

    if(argc == 1){
        help();
        return 0;
    }

    index = 1;
    while(index < argc){
        if( strcmp(argv[index], "3") == 0 ){
            force_irq = 3;
        }else if( strcmp(argv[index], "5") == 0 ){
            force_irq = 5;
        }else if( strcmp(argv[index], "7") == 0 ){
            force_irq = 7;
        }else if( strcmp(argv[index], "9") == 0 ){
            force_irq = 9;
        }else if( strcmp(argv[index], "10") == 0 ){
            force_irq = 10;
        }else if( strcmp(argv[index], "11") == 0 ){
            force_irq = 11;
        }else if( strcmp(argv[index], "edge") == 0 ){
            trigger_mode = 0;
        }else if( strcmp(argv[index], "EDGE") == 0 ){
            trigger_mode = 0;
        }else if( strcmp(argv[index], "level") == 0 ){
            trigger_mode = 1;
        }else if( strcmp(argv[index], "LEVEL") == 0 ){
            trigger_mode = 1;
        }
        index += 1;
    }

    YMFIRQ(force_irq);
    if(trigger_mode >= 0){
        set_irq_edge_or_level(force_irq, trigger_mode);
    }

    return 0;
}
